unit Deriver;

{******************************************************************************
 Unit Name:	 Deriver
 Description:Deriver defines TDeriver class that can be used to enable
						 polymorphic creation of objects based on general info.
 Author:     M. Venturini
 Revisor:
 Date:       17 Feb. 97
 Revision:
******************************************************************************}

interface

uses
	Classes, SysUtils;

type

  {****************************************************************************
   --	function type TMatchMethod [method]
   The class that are to be polymorphically create must provide this method
   to check if the creational info match the class requirements.
   --	Parameters:
  		Info:	      	in TObject
										Creational infos.
      Result:       out Boolean
      							True if the creational infos match the class requirements.
  ****************************************************************************}

	TMatchMethod = function (const Info: TObject): Boolean of object;

  TDeriver = class(TObject)
  private
  	InstanceCount: Integer;
  	Derived: TStringList;
  public
    procedure IncInstance;
    procedure DecInstance;
  	procedure Register(const DerivedClass: TClass;
    									 const MatchMethod: TMatchMethod);
  	function DerivedClass(const Info: TObject): TClass;
    constructor Create;
    destructor Destroy; override;
  end;

  EDeriver = class(Exception);
  EMatch = class(EDeriver);

{*****************************************************************************}

implementation

const
	EMatchMsg = ' - No derived class matched';

{* TDerived ******************************************************************}
{Derived is a list of TDerived istances that contains the class reference and
 the address of the method to be used to determine the class corresponding to
 a set of creational infos                                                    }

type
  TDerived = class(TObject)
  private
    DerivedClass: TClass;
    MatchMethod: TMatchMethod;
  public
  	constructor Create(const DerivedClass: TClass;
                       const MatchMethod: TMatchMethod);
  end;

constructor TDerived.Create(const DerivedClass: TClass;
										        const MatchMethod: TMatchMethod);
begin

	inherited Create;
  Self.DerivedClass := DerivedClass;
  Self.MatchMethod := MatchMethod;

end;

{* TDeriver ******************************************************************}

{******************************************************************************
 --	procedure IncInstance and DecInstance [method]
 These procedure mantain the InstanceCount, that is the number of objects that
 use the TDeriver object. When no object remain the TDeriver object can be
 safely destroyed.
 Note: these procedure are useless in Delphi 2.
******************************************************************************}

procedure TDeriver.IncInstance;
begin

	Inc(InstanceCount);

end;

procedure TDeriver.DecInstance;
begin

	Dec(InstanceCount);
  if InstanceCount = 0 then
  	Destroy;

end;

{******************************************************************************
 --	procedure Register [method]
 Register a class in Derived list.
 --	Parameters:
 		DerivedClass:   in TClass
    								Class to be registerd.
    MatchMethod:    in TMatchMethod
                    Method to be used to determine the class corresponding
                    to a set of creational infos.
******************************************************************************}

procedure TDeriver.Register(const DerivedClass: TClass;
														const MatchMethod: TMatchMethod);
begin

  Derived.AddObject(DerivedClass.ClassName,TDerived.Create(DerivedClass,MatchMethod));

end;

{******************************************************************************
 --	function DerivedClass [method]
 For every class registered in Derived list the MatchMethod is called with
 creational infos as parameter. When MatchMethod return true the corresponding
 class is returned.
 --	Parameters:
 		Info:	      		in TObject
										Creational infos.
 		Result:         out TClass
    								Class that matches creational infos.
 -- Exception:
 	  EMatch:					raised if no class match is found.
******************************************************************************}

function TDeriver.DerivedClass(const Info: TObject): TClass;

var
	I: Integer;

begin

	I := 0;
 	repeat
    with Derived.Objects[I] as TDerived do
	    if MatchMethod(Info) then begin
	    	Result :=	DerivedClass;
  	    Exit;
    	end;
    Inc(I);
 	until I >= Derived.Count;
  raise EMatch.Create(Derived.Strings[0] + EMatchMsg);

end;

{******************************************************************************
 --	procedure Create [constructor]
******************************************************************************}

constructor TDeriver.Create;
begin

	inherited Create;
  InstanceCount := 0;
	Derived := TStringList.Create;

end;

{******************************************************************************
 --	procedure Create [destructor]
 Free everything.
******************************************************************************}

destructor TDeriver.Destroy;

var
	I: Integer;

begin

	for I := 0 to Derived.Count -1 do
  	(Derived.Objects[I] as TDerived).Free;
  Derived.Free;
  inherited Destroy;

end;


end.
